#!/usr/bin/env python
# setup-windows.py - Build system for Ubuntu One Client package
#
# Copyright 2010-2012 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
"""setup-windows.py"""

import os
import shutil
import subprocess
import sys

from distutils import log
from distutils.cmd import Command
from distutils.core import setup

import conf

LOG_LEVEL = "DEBUG"
LOG_FILE_SIZE = "1000000"


# pylint is giving spurious reimport errors
# pylint: disable=W0404
# pylint is complaining about parent classes having too many methods
# pylint: disable=R0904
# pylint says "too many branches"
# pylint: disable=R0912
def get_py2exe_extension():
    """Return an extension class of py2exe."""
    # pylint: disable=F0401
    from py2exe.build_exe import py2exe as build_exe
    # pylint: enable=F0401

    # pylint: disable=E1101
    # pylint: disable=W0231
    class MediaCollector(build_exe):
        """Extension that copies lazr missing data."""

        def __init__(self, *args, **kwargs):
            """Create a new instance."""
            build_exe.__init__(self, *args, **kwargs)

        def _add_module_data(self, module_name):
            """Add the data from a given path."""
            # Create the media subdir where the
            # Python files are collected.
            media = module_name.replace('.', os.path.sep)
            full = os.path.join(self.collect_dir, media)
            if not os.path.exists(full):
                self.mkpath(full)

            # Copy the media files to the collection dir.
            # Also add the copied file to the list of compiled
            # files so it will be included in zipfile.
            module = __import__(module_name, None, None, [''])
            for path in module.__path__:
                for file_name in os.listdir(path):
                    file_name = os.path.join(path, file_name)
                    log.info('Copying file %s', file_name)
                    name = os.path.basename(file_name)
                    if not os.path.isdir(file_name):
                        self.copy_file(file_name, os.path.join(full, name))
                        self.compiled_files.append(os.path.join(media, name))
                    else:
                        self.copy_tree(file_name, os.path.join(full, name))

        def copy_extensions(self, extensions):
            """Copy the missing extensions."""
            build_exe.copy_extensions(self, extensions)
            for module in [
                'lazr.uri',
                'lazr.restfulclient',
                'lazr.authentication',
                'wadllib',
                'ubuntuone.storageprotocol',
                ]:
                self._add_module_data(module)

    # ModuleFinder can't handle runtime changes to __path__,
    # but win32com uses them
    try:
        # py2exe 0.6.4 introduced a replacement modulefinder.
        # This means we have to add package paths there, not to the built-in
        # one.  If this new modulefinder gets integrated into Python, then
        # we might be able to revert this some day.
        # if this doesn't work, try import modulefinder
        # pylint: disable=F0401
        try:
            import py2exe.mf as modulefinder
        except ImportError:
            import modulefinder
        import win32com
        # pylint: enable=F0401
        for path in win32com.__path__[1:]:
            modulefinder.AddPackagePath("win32com", path)
        for extra in ["win32com.shell"]:  # ,"win32com.mapi"
            __import__(extra)
            modules = sys.modules[extra]
            for path in modules.__path__[1:]:
                modulefinder.AddPackagePath(extra, path)
    except ImportError:
        # no build path setup, no worries.
        pass
    # pylint: enable=E1101
    return MediaCollector
# pylint: enable=W0404
# pylint: enable=R0904
# pylint: enable=R0912


def branch_and_merge(urls):
    """Branch and merge all requested branches."""
    project_name = urls[0].split(":")[1]
    print "Getting sources for:", project_name
    folder_name = os.path.join("sources", project_name)
    current_folder = os.getcwd()
    if os.path.isdir(folder_name):  # Folder exists
        shutil.rmtree(folder_name)
    subprocess.check_call(["bzr", "branch", urls[0], folder_name])
    os.chdir(folder_name)
    for url in urls[1:]:
        print "Merging:", url
        subprocess.check_call(["bzr", "merge", url])
        subprocess.check_call(["bzr", "commit", "-m", '"merged %s"' % url])
    os.chdir(current_folder)


# pylint is complaining about parent classes having too many methods
# pylint: disable=R0904
class FetchSources(Command):

    """Branch or pull the sources from all packages."""

    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        """Perform the command."""
        # Checkout sources
        for folder in ["sources", "bin", "installed", "data"]:
            if not os.path.isdir(folder):
                os.mkdir(folder)
        branch_and_merge(conf.U1_CLIENT_BRANCHES)
        branch_and_merge(conf.U1_CONTROL_PANEL_BRANCHES)
        branch_and_merge(conf.UBUNTU_SSO_BRANCHES)
        branch_and_merge(conf.U1_WINDOWS_INSTALLER_BRANCHES)
        branch_and_merge(conf.U1_STORAGE_PROTOCOL_BRANCHES)
# pylint: enable=R0904


# pylint is complaining about parent classes having too many methods
# pylint: disable=R0904
class PrepareSources(Command):

    """Manipulate the sources look as if they were installed."""

    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        """Perform the command."""
        # Take binaries
        shutil.copyfile(os.path.join(
            "sources", "ubuntuone-client", "bin", "ubuntuone-syncdaemon"),
            os.path.join("bin", "ubuntuone-syncdaemon"))
        shutil.copyfile(os.path.join(
            "sources", "ubuntuone-client", "bin", "u1sdtool"),
            os.path.join("bin", "u1sdtool"))
        shutil.copyfile(os.path.join("sources", "ubuntu-sso-client",
            "bin", "ubuntu-sso-login"),
            os.path.join("bin", "ubuntu-sso-login"))
        shutil.copyfile(os.path.join("sources",
            "ubuntuone-control-panel", "bin",
            "ubuntuone-control-panel-qt"),
            os.path.join("bin", "ubuntuone-control-panel-qt"))
        shutil.copyfile(os.path.join("sources", "ubuntuone-windows-installer",
            "bin", "ubuntuone-installer-qt"),
            os.path.join("bin", "ubuntuone-installer-qt"))

        # Remove "installed" copy
        try:
            shutil.rmtree(os.path.join("installed", "Lib"))
        except OSError:
            pass

        # Build UI files, copy packages.
        dest_dir = os.path.join(
            "..", "..", "installed", "Lib", "site-packages")
        dest_sso = os.path.join(dest_dir, "ubuntu_sso")
        start_dir = os.getcwd()
        os.chdir(os.path.join("sources", "ubuntu-sso-client"))
        os.system("python setup.py build")
        os.system("python setup.py install --prefix=%s" %
            os.path.join("..", "..", "installed"))
        shutil.copytree("ubuntu_sso", dest_sso)
        sso_revno = subprocess.Popen(
            ["bzr", "revno", "."],
            stdout=subprocess.PIPE).communicate()[0]
        os.chdir(start_dir)

        os.chdir(os.path.join("sources", "ubuntuone-client"))
        shutil.copy(os.path.join("windows", "clientdefs.py"),
            os.path.join("ubuntuone", "clientdefs.py"))
        dest_client = os.path.join(dest_dir, "ubuntuone")
        shutil.copytree("ubuntuone", dest_client)
        u1client_revno = subprocess.Popen(
            ["bzr", "revno", "."],
            stdout=subprocess.PIPE).communicate()[0]
        os.chdir(start_dir)

        os.chdir(os.path.join("sources", "ubuntuone-control-panel"))
        os.system("python setup.py build")
        # Copying by hand because the install is borked
        dest_cp = os.path.join(dest_dir, "ubuntuone", "controlpanel")
        shutil.copytree(os.path.join("ubuntuone", "controlpanel"),
            dest_cp)
        u1cp_revno = subprocess.Popen(
            ["bzr", "revno", "."],
            stdout=subprocess.PIPE).communicate()[0]
        os.chdir(start_dir)

        os.chdir(os.path.join("sources", "ubuntuone-windows-installer"))
        os.system("python setup.py build")
        # Copying by hand because the install is borked
        dest_installer = os.path.join(dest_dir, "ubuntuone_installer")
        shutil.copytree("ubuntuone_installer", dest_installer)
        u1inst_revno = subprocess.Popen(
            ["bzr", "revno", "."],
            stdout=subprocess.PIPE).communicate()[0]
        os.chdir(start_dir)

        # Create revno file
        with open(os.path.join("data", "revnos.txt"), "w+") as revnos:
            revnos.write("ubuntu-sso-client: %s"
                    "ubuntuone-client: %s"
                    "ubuntuone-control-panel: %s"
                    "ubuntuonw-windows-installer: %s" %
                    (sso_revno, u1client_revno, u1cp_revno, u1inst_revno))

        # Copy storage-protocol's pem files
        shutil.copyfile(os.path.join("sources",
            "ubuntuone-storage-protocol", "data",
            "UbuntuOne-Go_Daddy_CA.pem"),
            os.path.join("data", "UbuntuOne-Go_Daddy_CA.pem"))
        shutil.copyfile(os.path.join("sources",
            "ubuntuone-storage-protocol", "data",
            "UbuntuOne-Go_Daddy_Class_2_CA.pem"),
            os.path.join("data", "UbuntuOne-Go_Daddy_Class_2_CA.pem"))

        # Copy syncdaemon config data
        shutil.copyfile(os.path.join("sources",
            "ubuntuone-client", "data",
            "syncdaemon.conf"),
            os.path.join("data", "syncdaemon.conf"))
        logging_path = os.path.join("sources",
            "ubuntuone-client", "data",
            "logging.conf.in")
        with open(logging_path, "rb") as logconf:
            data = logconf.read()
        data = data.replace("@LOG_LEVEL@", LOG_LEVEL)
        data = data.replace("@LOG_FILE_SIZE@", LOG_FILE_SIZE)
        with open(os.path.join("data", "logging.conf"), "wb") as logconf:
            logconf.write(data)
# pylint: enable=R0904


if __name__ == "__main__":

    # pylint: disable=C0103
    # Add the installed copy to the PYTHONPATH
    sys.path.insert(0, os.path.join("installed", "Lib", "site-packages"))

    console = [
        {
            'script': 'bin/u1sdtool',
        },
    ]

    windows = [
        {
            'script': 'bin/ubuntuone-syncdaemon',
            'icon_resources': [(0, 'ubuntu_one.ico')]
        },
        {
            'script': 'bin/ubuntuone-control-panel-qt',
            'icon_resources': [(0, 'ubuntu_one.ico')]
        },
        {
            'script': 'bin/ubuntu-sso-login',
            'icon_resources': [(0, 'ubuntu_one.ico')]
        },
        {
            'script': 'bin/ubuntuone-installer-qt',
            'icon_resources': [(0, 'ubuntu_one.ico')]
        },
    ]

    if conf.SHOW_CMD == True:
        console.extend(windows)
        windows = []

    # Build bundles
    setup(
        name='ubuntuone',
        version='2.0.3',
        license='GPL v3',
        author='',
        author_email='',
        description='Ubuntu One',
        url='https://one.ubuntu.com',
        cmdclass={
            'py2exe': get_py2exe_extension(),
            'fetch': FetchSources,
            'prepare': PrepareSources,
        },
        console=console,
        windows=windows,
        options={
            'py2exe': {
                'includes': ['google.protobuf.descriptor',
                    'sip',
                    'twisted.web.resource',
                    'ubuntu_sso.qt',
                    'ubuntu_sso.qt.gui',
                    'PyQt4.QtNetwork',
                    'oauth',
                    'ubuntuone_installer',
                    'configglue',
                    'configglue.app',
                    'configglue.inischema',
                    'configglue.inischema.glue',
                    'configglue.inischema.parsers',
                    'comtypes',
                    ],
                'excludes': ['fsm'],
                "dll_excludes": [
                    "MSVCP90.dll",
                    "MSWSOCK.dll",
                    "mswsock.dll",
                    "powrprof.dll",
                    ],
            }})
