# -*- coding: utf-8 -*-
#
# Copyright 2011 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.

"""The user interface for the Ubuntu One Installer."""

import gettext

from twisted.internet import defer
from PyQt4 import QtGui, QtCore

from ubuntu_sso import networkstate
from ubuntu_sso.qt.gui import (
    SSOWizardPage,
    EmailVerificationPage,
    ErrorPage,
    ForgottenPasswordPage,
    ResetPasswordPage,
)
from ubuntu_sso.utils.ui import (
    SUCCESS,
)
from ubuntu_sso.qt.current_user_sign_in_ui import \
    Ui_CurrentUserSignInPage
from ubuntu_sso.qt.email_verification_ui import \
    Ui_EmailVerificationPage
from ubuntu_sso.qt.error_message_ui import Ui_ErrorPage
from ubuntu_sso.qt.success_message_ui import Ui_SuccessPage
from ubuntu_sso.qt.forgotten_password_ui import \
    Ui_ForgottenPasswordPage
from ubuntu_sso.qt.reset_password_ui import Ui_ResetPasswordPage
from ubuntu_sso.qt.controllers import (
    ChooseSignInController,
    EmailVerificationController,
    ErrorController,
    ResetPasswordController,
    SetUpAccountController,
    SuccessController,
)

from ubuntuone.controlpanel import backend
from ubuntuone.controlpanel.gui.qt.loadingoverlay import LoadingOverlay

from ubuntuone.platform.credentials import (
    APP_NAME,
    TC_URL,
)

# Module used to include the resources into this file
# pylint: disable=W0611
from ubuntuone_installer.gui.qt.ui import images_rc
# pylint: enable=W0611
from ubuntuone_installer.logger import setup_logging
from ubuntuone_installer.gui.qt.ui import (
    choose_sign_in_ui,
    congratulations_ui,
    license_ui,
    setup_account_ui,
)

from ubuntuone_installer.gui import qt
from ubuntuone_installer.gui.qt.are_you_sure import AreYouSure
from ubuntuone_installer.gui.qt.currentuser import CurrentUserController
from ubuntuone_installer.gui.qt.folders import FoldersPage
from ubuntuone_installer.gui.qt.forgotten import ForgottenPasswordController
from ubuntuone_installer.gui.qt.local_folders import LocalFoldersPage
from ubuntuone_installer.gui.qt.network_detection import NetworkDetectionPage
from ubuntuone_installer.gui.qt.preferences import PreferencesPage
from ubuntuone_installer.gui.qt.sync_now_or_later import SyncNowOrLaterPage
from ubuntuone_installer.gui.qt.setup_account import SetupAccountPage
from ubuntuone_installer.gui.qt.side_widget import SideWidget

_ = gettext.gettext
CONGRATULATIONS = _("Congratulations")
CREDENTIALS_ERROR_TITLE = _("Error getting credentials")
CREDENTIALS_ERROR = _('Application will close.\n\n%r')
SIGN_IN = _("Sign in to Ubuntu One")
SIGN_IN_SUBTITLE = _("Sign in with your existing Ubuntu One"
    " username and password.")
SKIP_TOUR = _("Skip tour, go to Dashboard")
START_SETUP = _("Start setup")
SUCCESSFUL_LOGIN = _("Login successful")
TITLE_STYLE = "<span style=\"font-size:24px\">%s</span>"
UPDATE_TITLE = _("Updates are available")
UPDATE_SOFTWARE = _("There is a new version of Ubuntu One available. "
                    "Do you want to perform an upgrade?")
REMOVE_OLD_BETA_TITLE = _("Remove old Beta")
REMOVE_OLD_BETA = _("The old Ubuntu One Beta is installed in your system. Do "
                    "you want to uninstall it?")
REMOVE_OLD_BETA_ERROR = _("There was an error uninstall the old beta. Please "
                          "do this step manually")
MIGRATE_OLD_BETA_DATA_TITLE = _("Migrate old data")
MIGRATE_OLD_BETA = _("Do you want to migrate the files from the old Ubuntu "
                     "One Beta location to the new one?")
MIGRATE_OLD_BETA_DATA_ERROR = _("There was an error migrating your data. "
                                "Please do this step manually.")

# Invalid name logger
# pylint: disable=C0103
logger = setup_logging('qt.gui')
# pylint: enable=C0103


def _ask_user_question(title, question):
    """Ask the user a Yes/No question."""
    result = QtGui.QMessageBox.question(None, title, question,
            QtGui.QMessageBox.Yes, QtGui.QMessageBox.No)
    return result == QtGui.QMessageBox.Yes


def user_wants_to_update():
    """Ask the user if he really wants to update the software."""
    return _ask_user_question(UPDATE_TITLE, UPDATE_SOFTWARE)


def user_wants_to_remove_old_beta():
    """Ask the user if he really wants to remove the old beta."""
    return _ask_user_question(REMOVE_OLD_BETA_TITLE, REMOVE_OLD_BETA)


def user_wants_to_migrate_data():
    """Ask the user if he really wants to migrate the data."""
    return _ask_user_question(MIGRATE_OLD_BETA_DATA_TITLE, MIGRATE_OLD_BETA)


def _warn_user(title, message):
    """Warn the user."""
    QtGui.QMessageBox.warning(None, title, message, QtGui.QMessageBox.Ok)


def warn_user_beta_not_removed():
    """Let the user know that the data was not migrated."""
    _warn_user(REMOVE_OLD_BETA_TITLE, REMOVE_OLD_BETA_ERROR)


def warn_user_data_not_migrated():
    """Let the user know that the data was not migrated."""
    _warn_user(MIGRATE_OLD_BETA_DATA_TITLE, MIGRATE_OLD_BETA_DATA_ERROR)


class LicensePage(SSOWizardPage):
    """Wizard Page that displays the license info and links to the GPL."""

    def __init__(self, parent=None):
        super(LicensePage, self).__init__(license_ui.Ui_Form(), None, parent)
        self.header.setVisible(False)
        self.ui.textBrowser.setHtml(qt.LICENSE_CONTENT)
        self.agree_button = None

    # Invalid names of Qt-inherited methods
    # pylint: disable=C0103

    def initializePage(self):
        """Setup UI details."""
        # Set the right texts and connections for buttons
        self.setButtonText(QtGui.QWizard.CustomButton1, _("Agree && continue"))
        self.setButtonText(QtGui.QWizard.CancelButton,
            _("Disagree && uninstall"))

        # This is just to catch an exception thrown when nothing
        # is connected to the signal. It's not an exceptional
        # condition at all.
        try:
            self.wizard().customButtonClicked.disconnect()
        except TypeError:
            pass

        self.wizard().setButtonLayout([
            QtGui.QWizard.CancelButton,
            QtGui.QWizard.BackButton,
            QtGui.QWizard.Stretch,
            QtGui.QWizard.CustomButton1])

        self.agree_button = self.wizard().button(QtGui.QWizard.CustomButton1)
        self.agree_button.setDefault(True)
        self.agree_button.style().unpolish(self.agree_button)
        self.agree_button.style().polish(self.agree_button)
        self.wizard().customButtonClicked.connect(self.check_connection)

    # pylint: disable=W0212
    @defer.inlineCallbacks
    def check_connection(self, button_id):
        """Decide next ID based on network detection."""
        if button_id == QtGui.QWizard.CustomButton1:
            connected = yield self.wizard().check_connection()
            if connected:
                self.next = self.wizard().SIGN_IN_PAGE_ID
            else:
                self.next = self.wizard()._next_id
            self.wizard().next()
    # pylint: enable=W0212

    def nextId(self):
        """Return the next page ID."""
        return self.next


class SignInPage(SSOWizardPage):
    """Wizard Page that lets the user Sign into Ubuntu One."""

    # Invalid names of Qt-inherited methods
    # pylint: disable=C0103

    def initializePage(self):
        """Setup UI details."""
        self.ui.cancel_button.setText(_("Close window and setup later"))
        self.ui.cancel_button.clicked.connect(self.wizard().reject)
        # Layout without custom button 1,
        # without finish button
        # without cancel
        self.wizard().setButtonLayout([
            QtGui.QWizard.Stretch,
            QtGui.QWizard.NextButton])

    def showEvent(self, event):
        """Set existing_account_button as default when the page is shown."""
        super(SignInPage, self).showEvent(event)
        self.ui.existing_account_button.setDefault(True)
        self.ui.existing_account_button.style().unpolish(
            self.ui.existing_account_button)
        self.ui.existing_account_button.style().polish(
            self.ui.existing_account_button)

    def nextId(self):
        """Provide the next id."""
        return self.next


class CurrentUserSignInPage(SSOWizardPage):
    """Wizard Page that lets a current user Sign into Ubuntu One."""

    # Invalid names of Qt-inherited methods
    # pylint: disable=C0103

    def nextId(self):
        """Provide the next id."""
        return self.next

    def initializePage(self):
        """Setup UI details."""
        self.setButtonText(QtGui.QWizard.CancelButton,
            _("Cancel"))
        # Layout without custom button 1,
        # without finish button
        self.wizard().setButtonLayout([
            QtGui.QWizard.BackButton,
            QtGui.QWizard.Stretch])

        # Set sign_in_button as default when the page is shown.
        self.ui.sign_in_button.setDefault(True)
        self.ui.sign_in_button.setEnabled(False)
        self.ui.sign_in_button.setProperty("DisabledState",
            not self.ui.sign_in_button.isEnabled())
        self.ui.sign_in_button.style().unpolish(
            self.ui.sign_in_button)
        self.ui.sign_in_button.style().polish(
            self.ui.sign_in_button)

    def cleanupPage(self):
        """Reset the state of the wizard if Verification code was visited."""
        self.wizard()._next_id = None


class SuccessPage(SSOWizardPage):
    """Shown after SSO login, before setup."""

    # Invalid names of Qt-inherited methods
    # pylint: disable=C0103

    def __init__(self, parent=None):
        super(SuccessPage, self).__init__(
            Ui_SuccessPage(),
            SuccessController(title=SUCCESSFUL_LOGIN), parent)
        self.backend = backend.ControlBackend()
        self.finish_button = None
        self.next_button = None
        self.ui.image_label.setPixmap(
            QtGui.QPixmap(':/win_installer_graphic.png'))

    def initializePage(self):
        """Setup UI details."""
        self.wizard().setButtonLayout([
            QtGui.QWizard.Stretch,
            QtGui.QWizard.FinishButton,
            QtGui.QWizard.NextButton,
            ])
        self.setFinalPage(True)
        self.wizard().setButtonText(QtGui.QWizard.FinishButton,
            SKIP_TOUR)
        self.wizard().setButtonText(QtGui.QWizard.NextButton,
            START_SETUP)
        self.finish_button = self.wizard().button(QtGui.QWizard.FinishButton)
        self.next_button = self.wizard().button(QtGui.QWizard.NextButton)
        self.finish_button.style().unpolish(self.finish_button)
        self.finish_button.style().polish(self.finish_button)
        self.next_button.setDefault(True)
        self.next_button.style().unpolish(self.next_button)
        self.next_button.style().polish(self.next_button)
        self.backend.connect_files()

        self.ui.success_message_body.setText(
            SUCCESS % {'app_name': self.wizard().app_name})

    def nextId(self):
        """Provide the next id."""
        return self.next


class CongratulationsPage(SSOWizardPage):
    """Final page of the wizard."""

    def __init__(self, parent=None):
        super(CongratulationsPage, self).__init__(
            congratulations_ui.Ui_Form(), None, parent)
        self.setFinalPage(True)
        self.setTitle(CONGRATULATIONS)

    # Invalid names of Qt-inherited methods
    # pylint: disable=C0103

    def initializePage(self):
        """Setup UI details."""
        # We need custom buttons
        self.wizard().setButtonText(QtGui.QWizard.FinishButton,
            _("Go to my Ubuntu One dashboard"))
        # Layout without custom button 1,
        self.wizard().setButtonLayout([
            QtGui.QWizard.Stretch,
            QtGui.QWizard.FinishButton])
        # Set FinishButton as default when the page is shown.
        self.wizard().button(QtGui.QWizard.FinishButton).setDefault(True)
        self.wizard().button(QtGui.QWizard.FinishButton).style().unpolish(
            self.wizard().button(QtGui.QWizard.FinishButton))
        self.wizard().button(QtGui.QWizard.FinishButton).style().polish(
            self.wizard().button(QtGui.QWizard.FinishButton))

        # Add syncdaemon to autostart
        qt.utils.add_syncdaemon_to_autostart()


class MainWindow(QtGui.QWizard):

    """The Main Window of the Installer wizard."""

    # Invalid constant names and Qt-inherited methods
    # pylint: disable=C0103

    LICENSE_PAGE = 0
    SIGNIN_PAGE = 1
    SYNC_NOW_OR_LATER_PAGE = 2
    CONGRATULATIONS_PAGE = 3

    # definition of the signals raised by the widget (for SSO)
    recoverableError = QtCore.pyqtSignal('QString', 'QString')
    loginSuccess = QtCore.pyqtSignal('QString', 'QString')
    registrationIncomplete = QtCore.pyqtSignal('QString', 'QString')
    registrationSuccess = QtCore.pyqtSignal('QString', 'QString')
    userCancellation = QtCore.pyqtSignal('QString')

    def __init__(self, close_callback=None, installing=False):
        """Initialize this instance."""
        # Used to decide the next page dynamically
        self._next_id = None

        # Data for SSO
        self.app_name = APP_NAME
        self.tc_url = TC_URL
        # help_text is not displayed in the UI from design
        self.help_text = ""

        super(MainWindow, self).__init__()
        # Set windows flags
        flags = self.windowFlags() & ~QtCore.Qt.WindowContextHelpButtonHint
        self.setWindowFlags(flags)

        self.setWindowTitle(APP_NAME)
        self.setMinimumHeight(538)
        self.setMinimumWidth(740)
        self.setWizardStyle(self.ModernStyle)
        self.close_callback = close_callback

        # Label for Form Errors
        self.form_errors_label = QtGui.QLabel()
        self.form_errors_label.setObjectName("form_errors")
        self.form_errors_label.hide()

        self.setSideWidget(SideWidget())
        self.overlay = LoadingOverlay(self)
        self.overlay.hide()

        self.setOption(self.NoBackButtonOnStartPage, True)

        # PyQt doesn't support the (int, page) version of addPage, so
        # Add the pages in the right order
        # pylint: disable=C0103
        self.license_page = LicensePage()
        self.LICENSE_PAGE = self.addPage(self.license_page)
        self.network_page = NetworkDetectionPage()

        #SSO Pages
        title_page = TITLE_STYLE % SIGN_IN
        self.sign_in_controller = ChooseSignInController(
            title=title_page)
        self.sign_in_page = SignInPage(
            ui=choose_sign_in_ui.Ui_ChooseSignInPage(),
            controller=self.sign_in_controller,
            parent=self)
        if not installing:
            self.sign_in_page.ui.message_label.setText("")

        self.setup_controller = SetUpAccountController(message_box=self)
        self.setup_account = SetupAccountPage(
            setup_account_ui.Ui_SetUpAccountPage(),
            self.setup_controller,
            parent=self)

        email_controller = EmailVerificationController(message_box=self)
        email_verification_page = Ui_EmailVerificationPage()
        self.email_verification = EmailVerificationPage(
                                                       email_verification_page,
                                                       email_controller)
        self.email_verification.next_button.clicked.connect(
            self.overlay.show)

        title_page = TITLE_STYLE % SIGN_IN
        self.current_user_controller = CurrentUserController(
            title=title_page, subtitle=SIGN_IN_SUBTITLE,
            message_box=self)
        self.current_user = CurrentUserSignInPage(Ui_CurrentUserSignInPage(),
                                                  self.current_user_controller,
                                                  parent=self)
        self.current_user.ui.sign_in_button.clicked.connect(
            self.overlay.show)

        self.success = SuccessPage(self)
        self.error_controller = ErrorController()
        self.error = ErrorPage(Ui_ErrorPage(), self.error_controller)

        self.forgotten_pwd_controller = ForgottenPasswordController(
            message_box=self)
        self.forgotten = ForgottenPasswordPage(Ui_ForgottenPasswordPage(),
                                               self.forgotten_pwd_controller,
                                               parent=self)
        self.forgotten.ui.send_button.clicked.connect(
            self.overlay.show)

        self.reset_password_controller = ResetPasswordController(
            message_box=self)
        self.reset_password = ResetPasswordPage(Ui_ResetPasswordPage(),
                                                self.reset_password_controller,
                                                parent=self)
        self.reset_password.ui.reset_password_button.clicked.connect(
            self.overlay.show)

        self.SIGN_IN_PAGE_ID = self.addPage(self.sign_in_page)
        self.NETWORK_DETECTION_PAGE_ID = self.addPage(self.network_page)
        self.setup_account_page_id = self.addPage(self.setup_account)
        self.email_verification_page_id = self.addPage(self.email_verification)
        self.current_user_page_id = self.addPage(self.current_user)
        self.SUCCESS_PAGE = self.addPage(self.success)
        self.ERROR_PAGE = self.addPage(self.error)
        self.forgotten_password_page_id = self.addPage(self.forgotten)
        self.reset_password_page_id = self.addPage(self.reset_password)

        self.loginSuccess.connect(self.login_success_slot)
        self.registrationIncomplete.connect(self.registration_incomplete)
        self.registrationSuccess.connect(self.registration_success_slot)

        # End of SSO pages

        self.SYNC_NOW_OR_LATER_PAGE = self.addPage(SyncNowOrLaterPage())
        self.folders_page = FoldersPage()
        self.folders_page_id = self.addPage(self.folders_page)
        self.local_folders_page = LocalFoldersPage()
        self.local_folders_page_id = self.addPage(self.local_folders_page)
        self.preferences_page = PreferencesPage()
        self.preferences_page_id = self.addPage(self.preferences_page)
        self.CONGRATULATIONS_PAGE = self.addPage(CongratulationsPage())

        self.currentIdChanged.connect(self.on_currentIdChanged)
        # Set Wizard buttons style
        self.button(QtGui.QWizard.NextButton).setDefault(True)

        self.set_start_page(installing)

    # Invalid name "closeEvent"
    # pylint: disable=C0103

    @defer.inlineCallbacks
    def set_start_page(self, installing):
        """Set the Wizard start page based in some conditions."""
        connected = yield self.check_connection()
        if installing:
            self.setStartId(self.LICENSE_PAGE)
        elif connected:
            self.setStartId(self.SIGNIN_PAGE)
        else:
            self.setStartId(self._next_id)
        self.restart()

    def critical(self, message, page=None):
        """Show a message at the bottom of the page on form errors."""
        self.overlay.hide()
        self.form_errors_label.setText(message)
        self.form_errors_label.show()
        if page is None and self.currentPage():
            self.currentPage().layout().addWidget(self.form_errors_label)
        elif page is not None:
            page.layout().addWidget(self.form_errors_label)

    def next(self):
        """Show the next page to display and remove the form errors label."""
        self.overlay.hide()
        self.form_errors_label.hide()
        if self.currentPage().layout():
            self.currentPage().layout().removeWidget(self.form_errors_label)
        self.form_errors_label.setParent(None)
        super(MainWindow, self).next()

    def nextId(self):
        """Selects which page to display next."""
        if self._next_id is not None:
            return self._next_id
        else:
            return QtGui.QWizard.nextId(self)

    @defer.inlineCallbacks
    def check_connection(self):
        """Return is machine is connected, False set Network Page as next."""
        # pylint: disable=W0703
        connected = yield networkstate.is_machine_connected()
        if not connected:
            self._next_id = self.NETWORK_DETECTION_PAGE_ID
        defer.returnValue(connected)
        # pylint: enable=W0703

    def login_success_slot(self):
        """Called on successful login."""
        self._next_id = self.SUCCESS_PAGE
        self.next()
        self._next_id = self.SYNC_NOW_OR_LATER_PAGE

    def registration_incomplete(self):
        """Called when registration wasn't completed."""
        self._next_id = self.email_verification_page_id
        self.next()
        self._next_id = self.SUCCESS_PAGE

    def registration_success_slot(self):
        """Called on successful registration."""
        self._next_id = self.SUCCESS_PAGE
        self.next()
        self._next_id = self.SYNC_NOW_OR_LATER_PAGE

    def done(self, result):
        """The main window is being closed, call any custom callback."""
        # Finished, not cancelled
        if result == 1:  # Cancelled
            qt.utils.start_control_panel()
        elif self.currentId() == self.LICENSE_PAGE:
            if AreYouSure(self).exec_() == QtGui.QDialog.Accepted:
                qt.utils.uninstall_application()
            else:
                return
        elif self.currentId() != self.CONGRATULATIONS_PAGE:
            if AreYouSure(self).exec_() == QtGui.QDialog.Rejected:
                return
        if self.close_callback is not None:
            self.close_callback()
        QtGui.QWizard.done(self, result)
        self.userCancellation.emit(self.app_name)

    def on_currentIdChanged(self, page_id):
        """The current page changed."""
        if page_id < self.SIGNIN_PAGE:
            self.sideWidget().ui.states_frame.hide()
            self.sideWidget().set_stage(0)
        elif page_id < self.SYNC_NOW_OR_LATER_PAGE:
            self.sideWidget().ui.states_frame.show()
            self.sideWidget().set_stage(1)
        elif page_id < self.CONGRATULATIONS_PAGE:
            self.sideWidget().set_stage(3)
        else:
            self.sideWidget().set_stage(4)

    # Invalid name "resizeEvent"
    # pylint: disable=C0103

    def resizeEvent(self, event):
        """Resize the overlay to fit all the widget."""
        QtGui.QWizard.resizeEvent(self, event)
        self.overlay.resize(event.size())
