# -*- coding: utf-8 -*-

# Authors: Alejandro J. Cura <alecu@canonical.com>
#
# Copyright 2011 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.

"""The test suite for the Qt UI of the Ubuntu One Installer."""

from PyQt4 import QtGui, QtCore
from twisted.internet import defer
from ubuntuone.controlpanel.gui.qt import loadingoverlay

from ubuntuone_installer.gui.tests import FakedObject
from ubuntuone_installer.tests import TestCase

# Attribute 'yyy' defined outside __init__, access to a protected member
# pylint: disable=W0201, W0212


NO_OP = lambda *args: None


class FakeUi(FakedObject):
    """A fake Ui object."""

    exposed_methods = ['setupUi']


class FakeSignal(object):

    """A fake PyQt signal."""

    def __init__(self, *args, **kwargs):
        """Initialize."""
        self.target = None

    def connect(self, target):
        """Fake connect."""
        self.target = target

    def disconnect(self, *args):
        """Fake disconnect."""
        self.target = None

    def emit(self, *args):
        """Fake emit."""
        if self.target:
            self.target(*args)


class FakeBackend(object):

    """Fake ControlBackend."""

    account_info_data = None
    volumes_info_data = None

    def __init__(self, *args, **kwargs):
        self.called = []
        self._called = self
        self.clear = lambda: None
        self.connect_files = lambda: self.called.append('connect_files')
        self._is_valid = True
        self.volume_setings_changes = []
        self.folders_created = []

    def account_info(self, *args):
        """Fake account info."""
        return defer.succeed(self.account_info_data)

    def volumes_info(self, *args):
        """Fake volumes info."""
        return defer.succeed(self.volumes_info_data)

    def validate_path_for_folder(self, path):
        """Fake folder validation."""
        return self._is_valid

    def change_volume_settings(self, *args):
        """Fake change volume settings."""
        self.volume_setings_changes.append(args)

    def create_folder(self, *args):
        """Fake folder creation."""
        self.folders_created.append(args)


class FakeFailingBackend(object):
    """Fake Control Panel backend that fails."""

    def account_info(self, *args):
        """Fake account info."""
        return defer.fail(Exception())

    def volumes_info(self, *args):
        """Fake account info."""
        return defer.fail(Exception())


class FakeController(object):

    """A fake controller for the tests."""

    is_fake = True
    set_next_validation = lambda *a, **kw: None

    def __init__(self, *args, **kwargs):
        self.args = (args, kwargs)

    # pylint: disable=C0103
    def setupUi(self, view):
        """Fake the setup."""
    # pylint: enable=C0103


class FakeOverlay(object):

    """A fake delay overlay."""

    def __init__(self, *args, **kwargs):
        """Initialize."""
        self.show_counter = 0
        self.hide_counter = 0
        self.args = (args, kwargs)

    def show(self):
        """Fake show."""
        self.show_counter += 1

    def hide(self):
        """Fake hide."""
        self.hide_counter += 1

    def resize(self, *args):
        """Fake resize."""


class FakeWizard(object):
    """Replace wizard() function on wizard pages."""

    customButtonClicked = QtCore.QObject()

    def __init__(self):
        self.overlay = FakeOverlay()
        self.called = []
        self.buttons = {}
        self._next_id = -1

    # Invalid name "setButtonLayout", "setOption"
    # pylint: disable=C0103

    def setButtonLayout(self, *args, **kwargs):
        """Fake the functionality of setButtonLayout on QWizard class."""
        self.called.append(('setButtonLayout', (args, kwargs)))

    def setOption(self, *args, **kwargs):
        """Fake the functionality of setOption on QWizard class."""
        self.called.append(('setOption', (args, kwargs)))

    # pylint: enable=C0103

    def button(self, button_id):
        """Fake the functionality of button on QWizard class."""
        return self.buttons.setdefault(button_id, QtGui.QPushButton())


class FakeWizardButtonStyle(FakeWizard):

    """Fake Wizard with button style implementation."""

    SIGN_IN_PAGE_ID = 1

    # pylint: disable=C0103
    def __init__(self):
        super(FakeWizardButtonStyle, self).__init__()
        self.data = {}
        self.customButtonClicked = self

    def setDefault(self, value):
        """Fake setDefault for button."""
        self.data['default'] = value

    def isDefault(self):
        """Fake isDefault."""
        return self.data['default']
    # pylint: enable=C0103

    def connect(self, func):
        """Fake customButtonClicked connect."""
        self.data['connect'] = func

    def disconnect(self, func):
        """Fake customButtonClicked disconnect."""
        self.data['disconnect'] = func

    def button(self, button_id):
        """Fake the functionality of button on QWizard class."""
        return self

    def style(self):
        """Fake button style."""
        return self

    def polish(self, button):
        """Fake polish."""
        self.data['polish'] = button

    def unpolish(self, button):
        """Fake unpolish."""
        self.data['unpolish'] = button

    def next(self):
        """Fake next for wizard."""
        self.data['next'] = True


class FakeWizardPage(object):

    """A fake wizard page."""

    def __init__(self, *args, **kwargs):
        self.has_back_button = True

    # pylint: disable=C0103
    def initializePage(self):
        """Fake initializePage."""


class BaseTestCase(TestCase):
    """Base Test Case."""

    class_ui = None
    kwargs = {}

    @defer.inlineCallbacks
    def setUp(self):
        yield super(BaseTestCase, self).setUp()
        # self.class_ui is not callable
        # pylint: disable=E1102
        # pylint: disable=C0103
        self.ui = None
        if self.class_ui is not None:
            self.ui = self.class_ui(**self.kwargs)

        if hasattr(self.ui, 'backend'):
            # clean backend calls
            self.ui.backend._called.clear()

        self.patch(loadingoverlay, 'LoadingOverlay', FakeOverlay)

    def get_pixmap_data(self, pixmap):
        """Get the raw data of a QPixmap."""
        byte_array = QtCore.QByteArray()
        array_buffer = QtCore.QBuffer(byte_array)
        pixmap.save(array_buffer, "PNG")
        return byte_array

    # Invalid name "assertEqualPixmap"
    # pylint: disable=C0103

    def assertEqualPixmaps(self, pixmap1, pixmap2):
        """Compare two Qt pixmaps."""
        d1 = self.get_pixmap_data(pixmap1)
        d2 = self.get_pixmap_data(pixmap2)
        self.assertEqual(d1, d2)
