# -*- coding: utf-8 -*-

# Authors: Diego Sarmentero <diego.sarmentero@canonical.com>
#
# Copyright 2011 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Utils Functions Platform specific."""

import sys

from ubuntuone.platform.credentials import CredentialsManagementTool
from twisted.internet import defer

from ubuntuone_installer.logger import setup_logging


logger = setup_logging('qt.utils')


# pylint: disable=C0103
if sys.platform == 'win32':
    from ubuntuone_installer.gui.qt.utils import windows
    uninstall_application = windows.uninstall_application
    start_control_panel = windows.start_control_panel
    add_syncdaemon_to_autostart = windows.add_syncdaemon_to_autostart
    default_folders = windows.default_folders
    is_old_beta_installed = windows.is_old_beta_installed
    are_updates_present = windows.are_updates_present
    perform_update = windows.perform_update
    uninstall_old_beta = windows.uninstall_old_beta
    migrate_old_data = windows.migrate_old_data
    check_updates = windows.check_updates
    perform_old_beta_data_migration = windows.perform_old_beta_data_migration
else:
    from ubuntuone_installer.gui.qt.utils import linux
    uninstall_application = linux.uninstall_application
    start_control_panel = linux.start_control_panel
    add_syncdaemon_to_autostart = linux.add_syncdaemon_to_autostart
    default_folders = linux.default_folders
    is_old_beta_installed = lambda: False
    are_updates_present = lambda *args, **kwargs: False
    perform_update = lambda *args, **kwargs: None
    uninstall_old_beta = lambda *args, **kwargs: None
    migrate_old_data = lambda *args, **kwargs: None
    check_updates = lambda *args, **kwargs: None
    perform_old_beta_data_migration = lambda *args, **kwargs: None
# pylint: enable=C0103


@defer.inlineCallbacks
def check_credentials(stop_cb, installing):
    """Check credentials and either start u1cp or show the wizard."""
    credtool = CredentialsManagementTool()
    try:
        creds = yield credtool.find_credentials()
    except:  # pylint: disable=W0702
        logger.exception('Error while getting the credentials:')
        stop_cb()
    else:
        from ubuntuone_installer.gui.qt import gui  # pylint: disable=W0404
        if creds:  # Have credentials already
            logger.info('Got credentials! starting controlpanel.')
            start_control_panel(with_icon=False)
            logger.info('Checking updates')
            yield check_updates(gui)
            logger.info('Updates checked, stopping.')
            stop_cb()
        else:  # No credentials
            # if there are no credentials it usually means that we
            # are installing the application, before we do anything
            # lets check if the old beta is present
            yield perform_old_beta_data_migration(gui)
            logger.info('Got empty credentials, starting wizard app.')
            window = gui.MainWindow(close_callback=stop_cb,
                                    installing=installing)
            window.show()
