# -*- coding: utf-8 -*-
#
# Copyright 2011 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for the multiplatform functions."""

import logging

from twisted.internet import defer
from ubuntuone.devtools.handlers import MementoHandler

from ubuntuone_installer.gui.qt import utils, gui
from ubuntuone_installer.gui.qt.tests import BaseTestCase


class FakeCredsManagementTool(object):
    """Fake a creds management tool."""

    raise_error = None
    credentials = None

    def find_credentials(self):
        """Fake find credentials."""
        if FakeCredsManagementTool.raise_error is not None:
            return defer.fail(FakeCredsManagementTool.raise_error)
        else:
            return defer.succeed(FakeCredsManagementTool.credentials)


class FakeWindow(object):
    """Fake a windows."""

    args = None
    visible = False

    def __init__(self, *a, **kw):
        FakeWindow.args = (a, kw)
        self.show = lambda: setattr(FakeWindow, 'visible', True)


class CheckCredentialsTestCase(BaseTestCase):
    """Test the code that is used to check the credentials."""

    @defer.inlineCallbacks
    def setUp(self):
        """Set the different tests."""
        yield super(CheckCredentialsTestCase, self).setUp()
        self.called = []
        self.patch(utils, 'CredentialsManagementTool', FakeCredsManagementTool)

        self.memento = MementoHandler()
        self.memento.setLevel(logging.DEBUG)
        utils.logger.addHandler(self.memento)
        self.addCleanup(utils.logger.removeHandler, self.memento)

        self.stop_cb = lambda: self.called.append('stop_cb')

        self.start_cp = lambda with_icon: \
                        self.called.extend(('start_cp', with_icon))

        self.patch(utils, 'start_control_panel', self.start_cp)

        self.check_updates = lambda _: self.called.append('check_updates')

        self.patch(utils, 'check_updates', self.check_updates)

        self.migrate_data_cb = lambda gui: \
                        self.called.extend(('migrate_data_cb', gui))

        self.patch(utils, 'perform_old_beta_data_migration',
           self.migrate_data_cb)

        self.patch(gui, 'MainWindow', FakeWindow)

    @defer.inlineCallbacks
    def test_check_credentials_failure(self):
        """Test when the credtool fail."""
        exc = ValueError('foo')
        self.patch(FakeCredsManagementTool, 'raise_error', exc)
        yield utils.check_credentials(stop_cb=self.stop_cb, installing=False)

        self.assertEqual(self.called, ['stop_cb'])
        self.assertTrue(self.memento.check_exception(exc.__class__, 'foo'))

    @defer.inlineCallbacks
    def test_check_credentials_present(self):
        """Test when creds are present."""
        self.patch(FakeCredsManagementTool, 'credentials', {'foo': 'bar'})
        yield utils.check_credentials(stop_cb=self.stop_cb, installing=False)

        self.assertEqual(self.called, ['start_cp', False,
                                       'check_updates', 'stop_cb'])
        msgs = ('Got credentials! starting controlpanel.', 'Checking updates',
                'Updates checked, stopping.')
        for msg in msgs:
            self.assertTrue(self.memento.check_info(msg))

    @defer.inlineCallbacks
    def test_check_credentials_not_present(self):
        """Test when creds are present."""
        self.patch(FakeCredsManagementTool, 'credentials', {})
        yield utils.check_credentials(stop_cb=self.stop_cb, installing=False)

        kw = dict(close_callback=self.stop_cb, installing=False)
        self.assertEqual(FakeWindow.args, ((), kw))
        self.assertTrue(FakeWindow.visible)
        msgs = ('Got empty credentials', 'starting wizard')
        self.assertTrue(self.memento.check_info(*msgs))
        self.assertEqual(self.called, ['migrate_data_cb', gui])
