# -*- coding: utf-8 -*-

# Authors: Diego Sarmentero <diego.sarmentero@canonical.com>
#          Manuel de la Pena <manuel@canonical.com>
#
# Copyright 2011 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for the platform specific functions."""

import sys
import os
import subprocess

# Avoid pylint error on Linux
# pylint: disable=F0401
import win32api
# pylint: enable=F0401

from twisted.internet import defer
# pylint: disable=E0611
from ubuntuone.platform import tools
# pylint: enable=E0611

from ubuntuone_installer.gui.qt import gui
from ubuntuone_installer.gui.qt import utils
from ubuntuone_installer.gui.qt.tests import BaseTestCase

# we want to test private methods
# pylint:disable=W0212


class UninstallerTestCase(BaseTestCase):

    """Test the uninstaller."""

    def test_execute_uninstall_on_licence_cancel_file_not_exists(self):
        """Pressing Disagree button from license page the uninstall is exec."""
        # If it's frozen this test fails because the path
        # depends on the location of the exe
        frozen = getattr(sys, "frozen", None)
        if frozen is not None:
            delattr(sys, "frozen")
            self.addCleanup(setattr, sys, 'frozen', frozen)
        self.patch(win32api, "ShellExecute", self._set_called)
        self.assertFalse(self._called)

    def test_execute_uninstall_on_licence_cancel_frozen_file_not_exists(self):
        """Pressing Disagree button from license page the uninstall is exec."""
        self.patch(win32api, "ShellExecute", self._set_called)
        sys.frozen = True
        self.addCleanup(delattr, sys, "frozen")
        utils.uninstall_application()
        self.assertFalse(self._called)

    def test_execute_uninstall_on_licence_cancel(self):
        """Pressing Disagree button from license page the uninstall is exec."""
        self.patch(win32api, "ShellExecute", self._set_called)
        self.patch(os.path, "exists", lambda path: True)
        utils.uninstall_application()
        # If it's frozen this test fails because the path
        # depends on the location of the exe
        frozen = getattr(sys, "frozen", None)
        if frozen is not None:
            delattr(sys, "frozen")
            self.addCleanup(setattr, sys, 'frozen', frozen)
        exec_path = os.path.abspath(__file__)
        folder_path_utils_dir = os.path.dirname(exec_path)
        folder_path_qt_dir = os.path.dirname(folder_path_utils_dir)
        folder = os.path.dirname(folder_path_qt_dir)
        uninstall_path = os.path.join(os.path.dirname(folder), "uninstall.exe")
        self.assertEqual(self._called,
            ((None, 'runas', uninstall_path, '--mode win32', '', 0), {}))

    def test_execute_uninstall_on_licence_cancel_frozen(self):
        """Pressing Disagree button from license page the uninstall is exec."""
        self.patch(win32api, "ShellExecute", self._set_called)
        self.patch(os.path, "exists", lambda path: True)
        sys.frozen = True
        self.addCleanup(delattr, sys, "frozen")
        utils.uninstall_application()
        self.assertTrue(self._called[0][0] is None)
        self.assertIn("uninstall.exe", self._called[0][2])
        self.assertEqual('runas', self._called[0][1])
        self.assertEqual('--mode win32', self._called[0][3])
        self.assertEqual('', self._called[0][4])
        self.assertEqual(0, self._called[0][5])


class StartControlPanelTestCase(BaseTestCase):

    """Test the start of the control panel."""

    def test_start_control_panel_with_icon(self):
        """The proper process is started."""
        self.patch(subprocess, "Popen", self._set_called)
        utils.start_control_panel()
        expected = ((['ubuntuone-control-panel-qt', '--with-icon'],), {})
        self.assertEqual(self._called, expected)

    def test_start_control_panel_without_icon(self):
        """The proper process is started."""
        self.patch(subprocess, "Popen", self._set_called)
        utils.start_control_panel(with_icon=False)
        expected = ((['ubuntuone-control-panel-qt'],), {})
        self.assertEqual(self._called, expected)

    def test_start_control_panel_frozen(self):
        """When frozen, the control-panel has a path."""
        # I can't patch sys because frozen is not there by default
        sys.frozen = True
        self.addCleanup(delattr, sys, 'frozen')

        root = 'Z:\\Yadda\\Yoo'
        self.patch(os.path, 'abspath', lambda p: os.path.join(root, p))
        self.patch(os.path, 'dirname', lambda p: p)
        exe = 'foo\\bar\\baz\\'
        self.patch(sys, 'executable', exe)

        self.patch(subprocess, "Popen", self._set_called)

        utils.start_control_panel()

        path = os.path.join(root, exe, 'ubuntuone-control-panel-qt.exe')
        expected = (([path, '--with-icon'],), {})
        self.assertEqual(self._called, expected)


class FakeRegistry(object):

    """A fake registry."""

    # pylint: disable=C0103
    HKEY_CURRENT_USER = 2
    KEY_ALL_ACCESS = 4
    REG_SZ = 1

    def __init__(self):
        self.HKEY_CURRENT_USER = 2
        self.KEY_ALL_ACCESS = 4
        self.connect_args = None
        self.openkey_args = None
        self.closekey_args = None
        self.query_args = None
        self.set_args = []

    def ConnectRegistry(self, *args, **kwargs):
        """Fake ConnectRegistry."""
        self.connect_args = (args, kwargs)
        return "REGISTRY"

    def OpenKey(self, *args, **kwargs):
        """Fake OpenKey."""
        self.openkey_args = (args, kwargs)
        return "KEY"

    def CloseKey(self, *args, **kwargs):
        """Fake CloseKey."""
        self.closekey_args = (args, kwargs)

    def QueryValueEx(self, *args, **kwargs):
        """Fake QueryValueEx."""
        self.query_args = (args, kwargs)

    def SetValueEx(self, *args, **kwargs):
        """Fake SetValueEx."""
        self.set_args.append((args, kwargs))


class AutostartTestCase(BaseTestCase):

    """Test add_syncdaemon_to_autostart."""

    @defer.inlineCallbacks
    def setUp(self):
        """Initialize this test instance."""
        self.registry = FakeRegistry()
        self.patch(utils.windows, "_winreg", self.registry)
        yield super(AutostartTestCase, self).setUp()

    def test_add_syncdaemon_to_autostart(self):
        """Check that the registry is updated correctly."""
        # I can't patch sys because frozen is not there by default
        sys.frozen = True
        self.addCleanup(delattr, sys, 'frozen')
        utils.windows.add_syncdaemon_to_autostart()
        self.assertEqual(self.registry.connect_args,
            ((None, self.registry.HKEY_CURRENT_USER), {}))
        self.assertEqual(self.registry.openkey_args,
            (('REGISTRY', utils.windows.AUTORUN_KEY, 0,
              self.registry.KEY_ALL_ACCESS), {}))
        self.assertEqual(self.registry.closekey_args, (('KEY',), {}))
        self.assertEqual(self.registry.query_args, None)
        self.assertEqual(self.registry.set_args,
             [(('KEY', 'Ubuntu One', 0, 1,
             '"C:\\Python27\\ubuntuone-syncdaemon.exe"'), {}),
             (('KEY', 'Ubuntu One Icon', 0, 1,
             '"C:\\Python27\\ubuntuone-control-panel-qt.exe" --minimized'),
             {})])

    def test_not_added_if_not_frozen(self):
        """Not frozen binaries are not added to the registry."""
        utils.windows.add_syncdaemon_to_autostart()
        self.assertEqual(self.registry.connect_args, None)
        self.assertEqual(self.registry.openkey_args, None)
        self.assertEqual(self.registry.closekey_args, None)
        self.assertEqual(self.registry.query_args, None)
        self.assertEqual(self.registry.set_args, [])


class DefaultFoldersTestCase(BaseTestCase):

    """Test the default suggested UDFs."""

    def test_special_folders(self):
        """Test that default_folders does the right thing."""
        folders = utils.default_folders()
        self.assertEqual(len(folders), 3)
        self.assertEqual(folders, [os.path.abspath(f) for f in folders])
        # pylint: disable=W0404
        import ctypes
        dll = ctypes.windll.shell32
        buf = ctypes.create_unicode_buffer(300)
        dll.SHGetSpecialFolderPathW(None, buf, utils.windows.CSIDL_PERSONAL,
                                    False)
        docs = buf.value
        dll.SHGetSpecialFolderPathW(None, buf, utils.windows.CSIDL_MYMUSIC,
                                    False)
        music = buf.value
        dll.SHGetSpecialFolderPathW(None, buf, utils.windows.CSIDL_MYPICTURES,
                                    False)
        pictures = buf.value
        expected = [docs, music, pictures]
        self.assertEqual(sorted(folders), sorted(expected))


class OldBetaUpdateTestCase(BaseTestCase):

    """Test the update from the old beta."""

    @defer.inlineCallbacks
    def setUp(self):
        """Set the different tests."""
        yield super(OldBetaUpdateTestCase, self).setUp()
        self.root_dir = 'Path\\to\\Ubuntu One'
        self.called = {}
        self.called_paths = []
        self.docs_dir = 'Path\\to\\docs'

        class FakeSDTool(object):
            """Fake sd tool."""

            # pylint:disable=E0213
            def get_root_dir(i_self):
                """Fake return root dir."""
                return self.root_dir
            # pylint:enable=E0213

        def fake_defer_to_thread(fn, old_path, new_path):
            """Fake defer to thread."""
            self.called_paths.append((old_path, new_path))

        def fake_get_default_folders():
            """Fake get docs dir."""
            return [self.docs_dir, None, None]

        self.patch(utils.windows, 'default_folders',
                   fake_get_default_folders)
        self.patch(tools, 'SyncDaemonTool',
                FakeSDTool)
        self.patch(utils.windows, 'deferToThread', fake_defer_to_thread)

    def _assert_is_old_beta_installed(self, return_value, expected_result):
        """Helper to test the is_old_beta_installed."""
        # vars used to know that we did indeed call the msi lib correctly

        def fake_get_product_property(uid, property_name):
            """Fake call to the msi lib."""
            self.called['uid'] = uid
            self.called['property_name'] = property_name
            return return_value

        self.patch(utils.windows, 'get_property_for_product',
                   fake_get_product_property)
        self.assertEqual(utils.is_old_beta_installed(), expected_result)
        # assert that we indeed called the ctypes with the correct data
        self.assertEqual(utils.windows.OLD_BETA_UID_KEY,
                         self.called['uid'])
        self.assertEqual(utils.windows.VERSION_STRING_PROPERTY,
                         self.called['property_name'])

    def test_is_old_beta_installed_true(self):
        """Test the function used to know if it is installed."""
        self._assert_is_old_beta_installed('SomeString', True)

    def test_is_old_beta_installed_false(self):
        """Test the function used to know if it is installed."""
        self._assert_is_old_beta_installed(None, False)

    def _assert_get_product_property(self, uid, property_name, return_code,
                                     expected_result):
        """Generic test for getting the property."""

        def fake_get_product_info(uid, property_name, property_buffer, size):
            """Fake call to the msi lib."""
            self.called['called_uid_buffer'] = uid
            self.called['called_property_name'] = property_name
            if expected_result:
                property_buffer.value = expected_result
                self.called['called_property_buffer'] = expected_result
            self.called['called_size'] = size
            return return_code

        self.patch(utils.windows.ctypes.windll.msi, 'MsiGetProductInfoW',
                   fake_get_product_info)
        self.assertEqual(expected_result,
                         utils.windows.get_property_for_product(uid,
                                                                property_name))
        self.assertEqual(uid, self.called['called_uid_buffer'], uid)
        self.assertEqual(property_name, self.called['called_property_name'])

    def test_get_product_property_present(self):
        """Test that we do indeed get the correct value."""
        uid = u'{som-uid}'
        property_name = u'VersionString'
        return_code = 0
        self._assert_get_product_property(uid, property_name, return_code,
                                          property_name)

    def test_get_product_property_not_present(self):
        """Test that we do get None."""
        uid = u'{som-uid}'
        property_name = u'VersionString'
        return_code = 1212
        self._assert_get_product_property(uid, property_name, return_code,
                                          None)

    def test_uninstall_present(self):
        """Test uninstalling when the msi is present."""
        path = 'path/to/msi'

        def fake_get_property_for_product(uid, property_name, buf_size=None):
            """A fake method."""
            self.called['property_uid'] = uid
            self.called['property_name'] = property_name
            return path

        def fake_msi_install_product(callable_fn, path, cmd):
            """A fake install function."""
            self.called['callable_fn'] = callable_fn
            self.called['install_path'] = path
            self.called['cmd'] = cmd
            return defer.succeed(0)

        self.patch(utils.windows, 'get_property_for_product',
                   fake_get_property_for_product)
        self.patch(utils.windows, 'deferToThread',
                   fake_msi_install_product)
        utils.uninstall_old_beta()
        self.assertEqual(self.called['property_uid'],
                         utils.windows.OLD_BETA_UID_KEY)
        self.assertEqual(self.called['property_name'], u'LocalPackage')
        self.assertEqual(utils.windows.ctypes.windll.msi.MsiInstallProductW,
                         self.called['callable_fn'])
        self.assertEqual(self.called['install_path'], path)
        self.assertEqual(self.called['cmd'], u'REMOVE=ALL')

    def test_uninstall_error(self):
        """Test uninstalling with an error."""
        path = 'path/to/somewhere'

        def fake_get_property_for_product(uid, property_name, buf_size=None):
            """A fake method."""
            self.called['property_uid'] = uid
            self.called['property_name'] = property_name
            return path

        def fake_msi_install_product(callable_fn, path, cmd):
            """A fake install function."""
            self.called['callable_fn'] = callable_fn
            self.called['install_path'] = path
            self.called['cmd'] = cmd
            return defer.succeed(1)

        self.patch(utils.windows, 'get_property_for_product',
                   fake_get_property_for_product)
        self.patch(utils.windows, 'deferToThread',
                   fake_msi_install_product)
        self.assertFailure(utils.uninstall_old_beta(),
                           utils.windows.UninstallException)
        self.assertEqual(self.called['property_uid'],
                         utils.windows.OLD_BETA_UID_KEY)
        self.assertEqual(self.called['property_name'], u'LocalPackage')
        self.assertEqual(utils.windows.ctypes.windll.msi.MsiInstallProductW,
                         self.called['callable_fn'])
        self.assertEqual(self.called['install_path'], path)
        self.assertEqual(self.called['cmd'], u'REMOVE=ALL')

    def test_uninstall_not_present(self):
        """Test uninstalling when the msi is missing."""

        def fake_get_property_for_product(uid, property_name):
            """A fake method."""
            self.called['uid'] = uid
            self.called['property_name'] = property_name
            return None

        self.patch(utils.windows, 'get_property_for_product',
                   fake_get_property_for_product)
        self.assertFailure(utils.uninstall_old_beta(),
                           utils.windows.UninstallException)
        self.assertEqual(self.called['uid'], utils.windows.OLD_BETA_UID_KEY)
        self.assertEqual(self.called['property_name'], u'LocalPackage')

    @defer.inlineCallbacks
    def test_migrate_data_present(self):
        """Test the migration of present data."""
        dirs = ['Music', 'Photos', 'Personal']
        # generate the expected moves
        expected_moves = []

        for path in dirs:
            old_full_path = utils.windows.LONG_PATH_PREFIX + \
                                os.path.join(self.docs_dir, 'Ubuntu One', path)
            new_full_path = utils.windows.LONG_PATH_PREFIX + \
                                os.path.join(self.root_dir, path)
            expected_moves.append((old_full_path, new_full_path))

        def fake_exists(path):
            """Return false only when we start with the destination."""
            return not path.startswith(utils.windows.LONG_PATH_PREFIX +
                                       self.root_dir)

        self.patch(os.path, 'exists', fake_exists)
        self.patch(os, 'listdir', lambda *args: dirs)
        yield utils.migrate_old_data()
        for path in expected_moves:
            self.assertTrue(path in self.called_paths)

    @defer.inlineCallbacks
    def test_migrate_data_not_present(self):
        """Test the migration of not present data."""
        dirs = ['Music', 'Photos', 'Personal']

        self.patch(os.path, 'exists', lambda *args: False)
        self.patch(os, 'listdir', lambda *args: dirs)
        yield utils.migrate_old_data()
        self.assertEqual(0, len(self.called_paths))

    def test_migrate_data_destination_present(self):
        """Test when files are already present in dst."""
        dirs = ['Music', 'Photos', 'Personal']

        # always return true will mean that when we try to find if the
        # destination is present we will also get True
        self.patch(os.path, 'exists', lambda *args: True)
        self.patch(os, 'listdir', lambda *args: dirs)
        # patch to later double check it was never called
        self.assertFailure(utils.migrate_old_data(),
                utils.windows.DataMigrationException)
        self.assertEqual(0, len(self.called_paths))


class AutoupdaterTestCase(BaseTestCase):
    """Test the code that is used for the auto update process."""

    @defer.inlineCallbacks
    def setUp(self):
        """Prepare for the diff tests."""
        yield super(AutoupdaterTestCase, self).setUp()
        self.auto_update_path = r'path\to\exe'
        self.return_from_call = 0
        self.command = None
        self.args = []

        def fake_execute_process(command, args=None, path=None):
            """Fake async process execution."""
            self.command = command
            self.args = args
            return self.return_from_call

        self.patch(utils.windows, 'getProcessValue', fake_execute_process)
        self.patch(utils.windows, '_get_update_path',
                   lambda: self.auto_update_path)

    @defer.inlineCallbacks
    def test_are_updates_present_true(self):
        """Test when updates are present."""
        # the idea is simple, set the value to be returned from
        # the fake call, assert that we get true and also that
        # we did use the correct parameters.
        self.return_from_call = 0
        are_present = yield utils.are_updates_present()
        self.assertTrue(are_present, 'Updates should be present.')
        # lets assert that we did use the correct args
        expected_args = ('--mode', 'unattended')
        self.assertEqual(expected_args, self.args)
        self.assertEqual(self.command, self.auto_update_path)

    @defer.inlineCallbacks
    def test_are_updates_present_false(self):
        """Test when updates are not present."""
        # similar to test_are_updates_present_true but with diff retcode
        self.return_from_call = 40
        are_present = yield utils.are_updates_present()
        self.assertFalse(are_present, 'Updates should NOT be present.')
        # lets assert that we did use the correct args
        expected_args = ('--mode', 'unattended')
        self.assertEqual(expected_args, self.args)
        self.assertEqual(self.command, self.auto_update_path)

    def test_perform_update(self):
        """Test the method that perfroms the update."""
        self.patch(utils.windows.win32api, 'ShellExecute', self._set_called)
        utils.perform_update()
        self.assertIn(self.auto_update_path, self._called[0][2])
        self.assertEqual('runas', self._called[0][1])
        self.assertEqual('--unattendedmodeui none', self._called[0][3])
        self.assertEqual('', self._called[0][4])
        self.assertEqual(0, self._called[0][5])


class UpdatePathTestCase(BaseTestCase):
    """Test getting the update path."""

    @defer.inlineCallbacks
    def setUp(self):
        """Set the different tests."""
        yield super(UpdatePathTestCase, self).setUp()
        self.module_path = os.path.join('.', 'root', 'exe', 'dist')
        self.module_path = os.path.abspath(self.module_path)
        self.patch(os.path, 'exists', lambda *args: True)

    def _clean_file(self, module, path):
        """Set back the correct path."""
        module.__file__ = path

    def _clean_frozen(self, old_executable):
        """Set back not to be frozen."""
        del sys.frozen
        sys.executable = old_executable

    def test_get_update_path_not_frozen(self):
        """Test we return the path to levels up to the dist."""
        old_utils_path = utils.windows.__file__
        utils.windows.__file__ = os.path.join(self.module_path, 'windows.py')
        self.addCleanup(self._clean_file, utils.windows, old_utils_path)
        expected_path = os.path.join('root',
                utils.windows.AUTO_UPDATE_EXE)
        expected_path = os.path.abspath(expected_path)
        path = utils.windows._get_update_path()
        self.assertEqual(expected_path, path)

    def test_get_update_path_frozen(self):
        """Test we return the path to levels up to the dist."""
        sys.frozen = True
        old_exe = sys.executable
        sys.executable = os.path.join(self.module_path, 'ubuntuone.exe')
        self.addCleanup(self._clean_frozen, old_exe)
        expected_path = os.path.join('root', 'exe',
                utils.windows.AUTO_UPDATE_EXE)
        expected_path = os.path.abspath(expected_path)
        path = utils.windows._get_update_path()
        self.assertEqual(expected_path, path)


class CheckUpdatesTestCase(BaseTestCase):
    """Test the code that is used for the auto update check process."""

    @defer.inlineCallbacks
    def setUp(self):
        """Set the different tests."""
        yield super(CheckUpdatesTestCase, self).setUp()
        self.are_present = False
        self.user_wants_update = False
        self.called = []

        def fake_are_present():
            """Fake are updates present."""
            self.called.append('are_updates_present')
            return self.are_present

        def fake_perform_update():
            """Fake perform update."""
            self.called.append('perform_update')

        def fake_user_wants_to_update():
            """Fake user_wants_to_update."""
            self.called.append('user_wants_to_update')
            return self.user_wants_update

        self.patch(utils.windows, 'are_updates_present', fake_are_present)
        self.patch(utils.windows, 'perform_update', fake_perform_update)
        self.patch(gui, 'user_wants_to_update', fake_user_wants_to_update)

    @defer.inlineCallbacks
    def test_check_updates_not_present(self):
        """Test the method that interacts with the user."""
        self.are_present = False
        yield utils.check_updates(gui)
        self.assertTrue('are_updates_present' in self.called)
        self.assertFalse('perform_update' in self.called)

    @defer.inlineCallbacks
    def test_check_updates_present_not_user(self):
        """Test the method that interacts with the user."""
        self.are_present = True
        self.user_wants_update = False
        yield utils.check_updates(gui)
        self.assertTrue('are_updates_present' in self.called)
        self.assertTrue('user_wants_to_update' in self.called)
        self.assertFalse('perform_update' in self.called)

    @defer.inlineCallbacks
    def test_check_updates_present_user(self):
        """Test the method that interacts with the user."""
        self.are_present = True
        self.user_wants_update = True
        yield utils.check_updates(gui)
        self.assertTrue('are_updates_present' in self.called)
        self.assertTrue('user_wants_to_update' in self.called)
        self.assertTrue('perform_update' in self.called)
